﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Microsoft.Build.Framework;
using Microsoft.M;
using Microsoft.M.Parser;
using Microsoft.M.Syntax;

namespace MSqlProvider
{
    public enum OperationType
    {
        Where,
        Identity,
        None
    }

    public class SqlGenerator : ISqlGenerator
    {
        internal SqlGenerator(TextReader reader)
        {
            try
            {
                SourceParser parser = new SourceParser();
                syntaxTerm = (CompilationUnit)parser.ParseTerm(reader, "buffer.m");

                List<BuildErrorEventArgs> errors = new List<BuildErrorEventArgs>();
                List<BuildWarningEventArgs> warnings = new List<BuildWarningEventArgs>();

                parser.ErrorRaised += delegate(object s, BuildErrorEventArgs e1)
                {
                    errors.Add(e1);
                };

                parser.WarningRaised += delegate(object s, BuildWarningEventArgs e1)
                {
                    warnings.Add(e1);
                };

                if (!parser.LastParseSuccessful)
                {
                    StringBuilder sb = new StringBuilder();
                    foreach (var error in errors)
                    {
                        sb.AppendFormat("Error ({0}, {1}: {2}", error.LineNumber, error.ColumnNumber, error.Message);
                        sb.AppendLine();
                    }
                    throw new Exception(sb.ToString());
                }
            }
            finally
            {
                reader.Close();
            }
        }
    
        public static SqlGenerator Load(TextReader reader)
        {
            return new SqlGenerator(reader);
        }

        private readonly CompilationUnit syntaxTerm;

        #region Implementation of ISqlGenerator

        /// <summary>
        /// this will generate an eqavalent sql statement from 
        /// m grammer.
        /// </summary>
        /// <returns></returns>
        public string Generate()
        {
            StringBuilder builder = new StringBuilder();

            foreach (IModuleDeclaration declaration in syntaxTerm.Modules)
            {
                builder.AppendSetExtactAbort();
                builder.AppendGo();

                builder.AppendBeginTransaction();
                builder.AppendGo();

                builder.AppendSetANSINulls();
                builder.AppendGo();

                builder.Append("create schema [" + declaration.Name.Value + "];");
                builder.AppendGo();
                builder.Append(ProcessMembers(declaration.Name.Value, declaration.Members));

                builder.AppendCommitTransaction();
                builder.AppendGo();
            }

            return builder.ToString();
        }

        private string ProcessMembers(string name, IEnumerable<IDeclaration> declarations)
        {
            StringBuilder builder = new StringBuilder();

            TypeDeclaration typeDeclaration = null;

            foreach (IDeclaration declaration in declarations)
            {
                if (declaration is TypeDeclaration)
                {
                    typeDeclaration = (TypeDeclaration) declaration;
                }
             
                if (declaration is ExtentDeclaration)
                {
                    builder.Append(ProcessExtent(name, typeDeclaration, (ExtentDeclaration) declaration));
                }
            }
            return builder.ToString();
        }

        private string ProcessExtent(string name, ITypeDeclaration typeDeclareation, IExtentDeclaration extentDeclaration)
        {
            StringBuilder builder = new StringBuilder();

            builder.AppendNewLine();

            builder.Append(ProcessParameterizedExpression(name, extentDeclaration.Name.Value, (ParameterizedExpression) typeDeclareation.Type));

            builder.AppendGo();

            const string format = "Insert into [{0}].[{1}] ({2}) values ({3});";

            GraphExpression graphExpression = (GraphExpression) extentDeclaration.InitialValue;
          
            if (extentDeclaration.InitialValue != null)
            {
                foreach (GraphExpression successor in graphExpression.Successors)
                {
                    string[] paramList = new string[successor.Successors.Count];
                    string[] valueList = new string[successor.Successors.Count];

                    int index = 0;

                    foreach (var expression in successor.Successors)
                    {
                        if (expression is GraphExpression)
                        {
                            GraphExpression node = (GraphExpression) expression;

                            Literal literal = (Literal) (node).Successors[0];

                            paramList[index] = node.Name.Value;
                            valueList[index] = ParseLiteral(literal);
                        }

                        index++;
                    }
                    builder.Append(string.Format(format, name, extentDeclaration.Name, string.Join(",", paramList), string.Join(",", valueList)));
                    builder.AppendNewLine();
                }
            }

            builder.AppendGo();

            return builder.ToString();
        }

        private string ParseLiteral(Literal literal)
        {
            switch (literal.Kind)
            {
                case LiteralKind.Text:
                    return "N'" + literal.Value + "'";
                case LiteralKind.Integer32:
                    return "" + literal.Value + "";
                case LiteralKind.Unsigned64:
                    return "" + literal.Value + "";
                case LiteralKind.DateTime:
                    return "'" + literal.Value + "'";

            }
            return literal.Value.ToString();
        }

        private string ProcessParameterizedExpression(string module, string extent, ParameterizedExpression expression)
        {
            StringBuilder builder = new StringBuilder();

            DeclarationReference operation = (DeclarationReference)expression.Operation;

            if (GerOperation(operation) == OperationType.Where)
            {
                builder.Append(string.Format("create table [{0}].[{1}]", module, extent));

                builder.AppendNewLine();

                builder.Append("(");

                builder.AppendNewLine();

                foreach (IExpression argument in expression.Arguments)
                {
                    if (argument is Microsoft.M.EntityType)
                    {
                        Microsoft.M.EntityType entityType = (Microsoft.M.EntityType) argument;

                        int index = 1;

                        foreach (IExtentDeclaration extentDeclaration in entityType.Members)
                        {
                            builder.Append(string.Format("[{0}] {1} not null", extentDeclaration.Name, GetDataType((DeclarationReference) extentDeclaration.Type)));

                            if (index < entityType.Members.Count)
                            {
                                builder.Append(",");
                            }

                            builder.AppendNewLine();

                            index++;
                        }

                        builder.AppendNewLine();
                    }
                    else if (argument is ParameterizedExpression)
                    {
                        builder.Append(ProcessParameterizedExpression(module, extent, (ParameterizedExpression) argument));
                    }
                }
                builder.Append(")");
            }
            else if (GerOperation(operation) == OperationType.Identity)
            {
                builder.Append(",");
                builder.AppendNewLine();

                foreach (DeclarationReference argument in expression.Arguments)
                {
                    builder.AppendIdentity(extent, argument.Name);
                }

                builder.AppendNewLine();
            }

            return builder.ToString();
        }

        private OperationType GerOperation(IDeclarationReference reference)
        {
            switch (reference.Name)
            {
                case "$Operator$Where":
                    return OperationType.Where;
                case "$Operator$Identity":
                    return OperationType.Identity;
            }

            return OperationType.None;
        }

        private string GetDataType(IDeclarationReference reference)
        {
            switch (reference.Name)
            {
                case "Integer32":
                    return "int";
                case "Text":
                    return "nvarchar(max)";
            }
            return "nvarchar(max)";
        }

        #endregion
    }
}
